// Copyright (C) 2018-2019  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// wcfs_virtmem provides wcfs client integrated with user-spce virtmem layer.
//
// XXX provides isolated view.

#include <golang/libgolang.h>
using namespace golang;

// XXX hack: C++ does not have __builtin_types_compatible_p, but CCAN configure
// thinks it does because CCAN is configired via C, not C++.
#include <config.h>
#undef  HAVE_BUILTIN_TYPES_COMPATIBLE_P
#define HAVE_BUILTIN_TYPES_COMPATIBLE_P 0

#include <wendelin/bigfile/virtmem.h>
#include <wendelin/bigfile/ram.h>
#include <wendelin/bug.h>

#include <string>
#include <vector>

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdint.h>
#include <inttypes.h>


#include "wcfs_misc.h"


using std::vector;

typedef uint64_t Tid;
typedef uint64_t Oid;

// TidHead is invalid Tid which is largest Tid value and means @head.
const Tid TidHead = -1ULL;

static string h(uint64_t v); // v -> 016x hex representation
#define h_(v)   (h(v).c_str())
static error mmap_zero_into_ro(void *addr, size_t size);

// XXX ok?
struct IContext {
    virtual chan<structZ>   done() = 0;
    virtual error           err() = 0;
};

#if 0
interface(Context) {
    ifunc(chan<structZ> done());
    ifunc(error         err());
};

I<io::Reader>(f)

// XXX wrap T* as IContext
template<typename T>
class Context : public IContext {
    T *obj;
public:
    Context(T *obj) : obj(obj) {}

    chan<structZ>   done()  { return obj->done();   }
    error           err()   { return obj->err();    }
};
#endif

struct Conn;
struct _File;
struct _Mapping;
struct WatchLink;
struct SrvReq;

// WCFS represents filesystem-level connection to wcfs server.
// XXX doc
struct WCFS {
    string  mountpoint;

    tuple<Conn*, error>         connect(Tid at);
    string                      _path(const string &obj);
    tuple<os::File, error>      _open(const string &path, int flags=O_RDONLY);
    tuple<WatchLink*, error>    _openwatch();
};

// Conn represents logical connection that provides view of data on wcfs
// filesystem as of particular database state.
//
// XXX doc
struct Conn {
    WCFS        *_wc;
    Tid         at;
    WatchLink   *_wlink;

    sync::Mutex        _filemu;
    dict<Oid, _File*>  _filetab; // {} foid -> _file

private:
    void _pinner(IContext *ctx);
    void _pin1(SrvReq *req);
};

// _File represent isolated file view under Conn.
//
// XXX doc      XXX naming -> _FileView ?
struct _File {
    Conn        *wconn;
    Oid         foid;       // hex of ZBigFile root object ID
    size_t      blksize;    // block size of this file
    os::File    headf;      // file object of head/file
    // .headfsize    head/file size is known to be at least headfsize (size ↑=)

    dict<int64_t, Tid>  pinned; // {} blk -> rev   that wcfs already sent us for this file
    vector<_Mapping*>   mmaps;  // []_Mapping ↑blk_start    mappings of this file
};

// _Mapping represents one mapping of _File.
struct _Mapping {
    _File   *file;
    int     blk_start;  // offset of this mapping in file

    BigFileH *fileh;        // mmapped under this file handle
    uint8_t  *mem_start;    // mmapped memory [mem_start, mem_stop)
    uint8_t  *mem_stop;

    int64_t blk_stop() const {
        ASSERT((mem_stop - mem_start) % file->blksize == 0);
        return blk_start + (mem_stop - mem_start) / file->blksize;
    }

    error _remmapblk(int64_t blk, Tid at);
};


// StreamID stands for ID of a stream multiplexed over WatchLink.
typedef uint64_t StreamID;

// rxPkt internally represents data of one message received over WatchLink.
struct rxPkt {
    // stream over which the data was received
    StreamID stream;

    // raw data received/to-be-sent.
    // XXX not e.g. string as chan<T> currently does not support types with
    //     non-trivial copy. Note: we anyway need to limit rx line length to
    //     avoid DoS, but just for DoS the limit would be higher.
    uint16_t datalen;
    char     data[128 - sizeof(StreamID) - sizeof(uint16_t)];

    error  from_string(const string& rx);
    string to_string() const;
};
static_assert(sizeof(rxPkt) == 128);


// WatchLink represents /head/watch link opened on wcfs.
//
// It is created by WCFS::_openwatch().
//
// .sendReq()/.recvReq() provides raw IO in terms of wcfs invalidation protocol messages.
// .close() closes the link.
//
// It is safe to use WatchLink from multiple threads simultaneously.
class WatchLink {
    WCFS            *_wc;
    os::File        _f;      // head/watch file handle
    string          _rxbuf;  // buffer for read data from _f
    chan<structZ>   _rx_eof; // becomes ready when wcfs closes its tx side

    // inv.protocol message IO
    chan<rxPkt>     _acceptq;   // server originated messages go here
    sync::Mutex     _rxmu;
    bool            _rxdown;
    dict<StreamID, chan<rxPkt>>
                    _rxtab;   // {} stream -> rxq    server replies go via here
    set<StreamID>   _accepted;  // streams we accepted but did not replied yet

    StreamID        _req_next;  // stream ID for next client-originated request XXX -> atomic
    sync::Mutex     _txmu;      // serializes writes
    sync::Once      _txclose1;

#if 0
    func() _serveCancel
    sync.WorkGroup *_serveWG
#endif

public:
    friend tuple<WatchLink*, error> WCFS::_openwatch();
    error close();
    error recvReq(IContext *ctx, SrvReq *rx_into);
    tuple<string, error> sendReq(IContext *ctx, const string &req);

private:
    void _closeTX();
    error _serveRX(IContext *ctx);
    tuple<string, error> _readline();
    error _send(StreamID stream, const string &msg);
    error _write(const string &pkt);
    tuple<chan<rxPkt>, error> _sendReq(IContext *ctx, const string &req);
};

// SrvReq represents 1 server-initiated wcfs request received over /head/watch link.
// XXX -> PinReq?
struct SrvReq {
    StreamID    stream; // request was received with this stream ID
    Oid         foid;   // request is about this file
    int64_t     blk;    // ----//---- about this block
    Tid         at;     // pin to this at;  TidHead means unpin to head
};


// connect creates new Conn viewing WCFS state as of @at.
tuple<Conn*, error> WCFS::connect(Tid at) {
    WCFS *wc = this;
    // XXX err ctx

    WatchLink *wlink;
    error err;
    tie(wlink, err) = wc->_openwatch();
    if (err != nil)
        return make_tuple((Conn *)NULL, err);

    Conn *wconn = new Conn();
    // TODO support !isolated mode
    wconn->_wc      = wc;
    wconn->at       = at;
    wconn->_wlink   = wlink;


    // XXX reenable
#if 0
    pinCtx, wconn._pinCancel = context.with_cancel(context.background())
    wconn._pinWG = sync.WorkGroup(pinCtx)
    wconn._pinWG.go(wconn._pinner)
#endif

    return make_tuple(wconn, nil);
}

// XXX Conn::close
#if 0
# close releases resources associated with wconn.
# XXX what happens to file mmappings?
@func(Conn)
def close(wconn):
    wconn._wlink.close()
    wconn._pinCancel()
    try:
        wconn._pinWG.wait()
    except Exception as e:  # canceled - ok
        if e is not context.canceled:
            raise

    # close all files - both that have no mappings and that still have opened mappings.
    # XXX after file is closed mappings continue to survive, but we can no
    # longer maintain consistent view.
    with wconn._filemu:
        for f in wconn._filetab.values():
            f.headf.close()
            f.headf = None

            # XXX stop watching f

        wconn._filetab = None
#endif

// _pinner receives pin messages from wcfs and adjusts wconn mappings.
void Conn::_pinner(IContext *ctx) {
    Conn &wconn = *this;

    // XXX panic/exc -> log.CRITICAL

    while (1) {
        // XXX -> recv inplace into on-stack req ?
        // XXX -> err, handle EOF, abort on *
        SrvReq *req = wconn._wlink->recvReq(ctx);
        if (req == NULL)
            return  // XXX ok? (EOF - when wcfs closes wlink)

        // we received request to pin/unpin file block. handle it
        wconn._pin1(req);

        // XXX free req?
    }
}

// pin1 handles one pin request received from wcfs.
void Conn::_pin1(SrvReq *req) {
    Conn &wconn = *this;

    // XXX defer: reply either ack or nak on error
    // XXX return error?

    _File *f;
    bool  ok;
    wconn._filemu.lock();
    tie(f, ok) = wconn._filetab.get(req->foid);
    if (!ok) {
        wconn._filemu.unlock();
        // XXX err = we are not watching the file - why wcfs sent us this update?
    }


    // XXX relock wconn -> f

    for (auto mmap : f->mmaps) {    // XXX use ↑blk_start for binary search
        if (!(mmap->blk_start <= req->blk && req->blk < mmap->blk_stop()))
            continue;   // blk ∉ mmap

        // XXX reenable
        //trace("\tremmapblk %d @%s" % (req->blk, (h(req.at) if req.at else "head")))
        // check if virtmem did not dirtied page corresponding to this block already
        virt_lock();
        TODO (mmap->file->blksize != mmap->fileh->ramh->ram->pagesize);
        if (!__fileh_page_isdirty(mmap->fileh, req->blk))
            mmap->_remmapblk(req->blk, req->at);    // XXX err
        virt_unlock();
        //trace("\t-> remmaped");   XXX
    }

    // update f.pinned
    // XXX do it before ^^^ remmapblk (so that e.g. concurrent
    // discard/writeout see correct f.pinned) ?
    if (req->at == TidHead) {
        f->pinned.erase(req->blk);      // unpin to @head
    }
    else {
        f->pinned[req->blk] = req->at;
    }

    wconn._filemu.unlock();
}

// XXX Conn::mmap

// XXX Conn::resync

// _remmapblk remmaps mapping memory for file[blk] to be viewing database as of @at state.
//
// at=TidHead means unpin to head/ .
// NOTE this does not check whether virtmem already mapped blk as RW.
error _Mapping::_remmapblk(int64_t blk, Tid at) {
    // XXX err context? blk #<blk> @<at>
    _Mapping *mmap = this;

    ASSERT(mmap->blk_start <= blk && blk < mmap->blk_stop());
    _File *f = mmap->file;
    error err;

    uint8_t *blkmem = mmap->mem_start + (blk - mmap->blk_start)*f->blksize;
    os::File fsfile;
    bool fclose = false;
    if (at == TidHead) {
        fsfile = f->headf;
    }
    else {
        // TODO share @rev fd until wconn is resynced?
        tie(fsfile, err) = f->wconn->_wc->_open(
                fmt::sprintf("@%s/bigfile/%s", h_(at), h_(f->foid)));
        if (err != nil)
            return err;
        fclose = true;
    }
    defer([&]() {
        if (fclose)
            fsfile.close();
    });

    struct stat st;
    err = fsfile.stat(&st);
    if (err != nil)
        return err;
    if ((size_t)st.st_blksize != f->blksize)
        return fmt::errorf("wcfs bug: blksize changed: expected %lu; got %ld",
                f->blksize, st.st_blksize);

    // block is beyond file size - mmap with zeros - else access to memory
    // after file.size will raise SIGBUS. (assumes head/f size ↑=)
    if ((blk+1)*f->blksize > (size_t)st.st_size) {
        err = mmap_zero_into_ro(blkmem, 1*f->blksize);
        if (err != nil)
            return err;
    }
    // block is inside file - mmap file data
    else {
        err = mm::map_into(blkmem, 1*f->blksize, PROT_READ, MAP_SHARED, fsfile, blk*f->blksize);
        if (err != nil)
            return err;
    }

    return nil;
}

// XXX _Mapping::remmap_blk

// XXX _Mapping::unmap


// ---- WatchLink ----

// _openwatch opens new watch link on wcfs.
tuple<WatchLink*, error> WCFS::_openwatch() {
    WCFS *wc = this;

    // head/watch handle.
    os::File f;
    error err;
    tie(f, err) = wc->_open("head/watch", O_RDWR);
    if (err != nil)
        return make_tuple((WatchLink *)NULL, err);

    WatchLink *wlink = new(WatchLink);
    wlink->_wc        = wc;
    wlink->_f         = f;
    wlink->_rx_eof    = makechan<structZ>();
    wlink->_acceptq   = makechan<rxPkt>();
    wlink->_rxdown    = false;
    wlink->_req_next  = 1;

#if 0
    serveCtx, wlink._serveCancel = context.with_cancel(context.background())
    wlink->_serveWG = sync.WorkGroup(serveCtx)
    wlink->_serveWG.go(wlink._serveRX)
#endif

    return make_tuple(wlink, nil);
}

void WatchLink::_closeTX() {
    WatchLink &wlink = *this;
    wlink._txclose1.do_([&]() {
        // ask wcfs to close its tx & rx sides; close(wcfs.tx) wakes up
        // _serveRX on client (= on us). The connection can be already closed
        // by wcfs - so ignore errors when sending bye.
        (void)wlink._send(1, "bye");    // XXX stream ok?
        // XXX vvv should be ~ shutdown(TX, wlink._f), however shutdown does
        // not work for non-socket file descriptors. And even if we dup link
        // fd, and close only one used for TX, peer's RX will still be blocked
        // ad fds are referring to one file object which stays in opened
        // state. So just use ^^^ "bye" as "TX closed" message.
        // wlink._wtx.close();
    });
}

// close closes the link.
error WatchLink::close() {
    WatchLink& wlink = *this;

    wlink._closeTX();
#if 0
    wlink._serveCancel();
    // XXX we can get stuck here if wcfs does not behave as we want.
    // XXX in particular if there is a silly - e.g. syntax or type error in
    //     test code - we currently get stuck here.
    //
    // XXX -> better pthread_kill(SIGINT) instead of relying on wcfs proper behaviour?
    // XXX -> we now have `kill -QUIT` to wcfs.go on test timeout - remove ^^^ comments?
    error err = wlink._serveWG.wait();
    // canceled is expected and ok
    if (err == context.canceled)
        err = nil;
#else
    error err = nil;
#endif

    error err2 = wlink._f.close();
    if (err == nil)
        err = err2;

    return err;
}

// _serveRX receives messages from ._f and dispatches them according to streamID.
error WatchLink::_serveRX(IContext *ctx) {    // XXX error -> where ?
    WatchLink& wlink = *this;

    // when finishing - wakeup everyone waiting for rx
    defer([&]() {
        wlink._acceptq.close();
        wlink._rxmu.lock();
        wlink._rxdown = true;   // don't allow new rxtab registers
        wlink._rxmu.unlock();
        for (auto rxqi : wlink._rxtab) {
            auto rxq = rxqi.second;
            rxq.close();
        }
    });

    string l;
    error  err;
    rxPkt  pkt;

    while (1) {
        // NOTE: .close() makes sure .f.read*() will wake up
        tie(l, err) = wlink._readline();    // XXX +maxlen
        if (err == io::EOF_) {  // peer closed its tx
            // XXX what happens on other errors?
            wlink._rx_eof.close();
        }
        if (err != nil)
            return err;
        printf("C: watch  : rx: %s", l.c_str());

        err = pkt.from_string(l);
        if (err != nil)
            return err;

        if (pkt.stream == 0) { // control/fatal message from wcfs
            // XXX print -> receive somewhere?   XXX -> recvCtl ?
            printf("C: watch  : rx fatal: %s\n", l.c_str());
            //wlink.fatalv.append(msg);
            continue;
        }

        bool reply = (pkt.stream % 2 != 0);
        if (reply) {
            chan<rxPkt> rxq;
            bool ok;

            wlink._rxmu.lock();
            tie(rxq, ok) = wlink._rxtab.pop(pkt.stream);
            wlink._rxmu.unlock();
            if (!ok) {
                // wcfs sent reply on unexpected stream
                // XXX log + dowmn.
                printf("wcfs sent reply on unexpected stream\n");
                continue;
            }
            int _ = select({
                ctx->done().recvs(),    // 0
                rxq.sends(&pkt),        // 1
            });
            if (_ == 0)
                return ctx->err();
        }
        else {
            wlink._rxmu.lock();
                if (wlink._accepted.has(pkt.stream)) {
                    wlink._rxmu.unlock();
                    // XXX log + down
                    printf("wcfs sent request on already used stream\n");
                    continue;
                }
                // XXX clear _accepted not to leak memory after reply is sent?
                wlink._accepted.insert(pkt.stream);
            wlink._rxmu.unlock();
            int _ = select({
                ctx->done().recvs(),            // 0
                wlink._acceptq.sends(&pkt),     // 1
            });
            if (_ == 0)
                return ctx->err();
        }
    }
}

// _send sends raw message via specified stream.
//
// multiple _send can be called in parallel - _send serializes writes.
// XXX +ctx?
error WatchLink::_send(StreamID stream, const string &msg) {
    WatchLink *wlink = this;
    if (msg.find('\n') != string::npos)
        panic("msg has \\n");
    string pkt = fmt::sprintf("%lu %s\n", stream, msg.c_str());
    return wlink->_write(pkt);
}

//error WatchLink::_write(const Pkt *pkt) {
error WatchLink::_write(const string &pkt) {
    WatchLink *wlink = this;

    wlink->_txmu.lock();
    defer([&]() {
        wlink->_txmu.unlock();
    });

    //printf('C: watch  : tx: %r' % pkt)
    int n;
    error err;
    tie(n, err) = wlink->_f.write(pkt.c_str(), pkt.size());
    return err;
}

// sendReq sends client -> server request and returns server reply.
// XXX -> reply | None when EOF
tuple<string, error> WatchLink::sendReq(IContext *ctx, const string &req) {
    WatchLink *wlink = this;
    // XXX err ctx

    rxPkt       rx;
    chan<rxPkt> rxq;
    error       err;
    tie(rxq, err) = wlink->_sendReq(ctx, req);
    // XXX err

    int _ = select({
        ctx->done().recvs(),    // 0
        rxq.recvs(&rx),         // 1
    });
    if (_ == 0)
        return make_tuple("", ctx->err());

    // XXX check for EOF
    string reply = rx.to_string();
    return make_tuple(reply, nil);
}

tuple</*rxq*/chan<rxPkt>, error> WatchLink::_sendReq(IContext *ctx, const string &req) {
    WatchLink *wlink = this;
    // XXX err ctx?

    wlink->_txmu.lock(); // XXX -> atomic (currently uses arbitrary lock)
        StreamID stream = wlink->_req_next;
        wlink->_req_next = (wlink->_req_next + 2); // wraparound at uint64 max
    wlink->_txmu.unlock();

    auto rxq = makechan<rxPkt>(1);
    wlink->_rxmu.lock();
        if (wlink->_rxdown) {
            wlink->_rxmu.unlock();
            return make_tuple(nil, fmt::errorf("link is down"));
        }
        if (wlink->_rxtab.has(stream)) {
            wlink->_rxmu.unlock();
            panic("BUG: to-be-sent stream is present in rxtab");
        }
        wlink->_rxtab[stream] = rxq;
    wlink->_rxmu.unlock();

    error err = wlink->_send(stream, req);
    if (err != nil) {
        // remove rxq from rxtab
        wlink->_rxmu.lock();
        wlink->_rxtab.erase(stream);
        wlink->_rxmu.unlock();
        // no need to drain rxq - it was created with cap=1

        rxq = nil;
    }

    return make_tuple(rxq, err);
}

// recvReq receives client <- server request.
error WatchLink::recvReq(IContext *ctx, SrvReq *rx) {
    WatchLink& wlink = *this;

    rxPkt pkt;
    _ = select(
        ctx->done().recvs(),            // 0
        wlink._acceptq.recvs(&pkt),     // 1
    )
    if (_ == 0)
        return ctx.err();

    rx = _rx
    if rx is None:  // XXX recheck _serveRX vs EOF signalling
        return rx

    rx.stream = pkt.stream;
    pkt.to_sting();
    stream, msg = rx
    return SrvReq(wlink, stream, msg)
}

// _readline reads next raw line sent from wcfs.
tuple<string, error> WatchLink::_readline() {
    WatchLink& wlink = *this;
    char buf[128];

    size_t nl_searchfrom = 0;
    while (1) {
        auto nl = wlink._rxbuf.find('\n', nl_searchfrom);
        if (nl != string::npos) {
            auto line = wlink._rxbuf.substr(0, nl+1);
            wlink._rxbuf = wlink._rxbuf.substr(nl+1);
            return make_tuple(line, nil);
        }
        nl_searchfrom = wlink._rxbuf.length();

        int n;
        error err;
        tie(n, err) = wlink._f.read(buf, sizeof(buf));
        if (n > 0) {
            // XXX limit line length to avoid DoS
            wlink._rxbuf += string(buf, n);
            continue;
        }
        if (err == nil)
            panic("read returned (0, nil)");
        if (err == io::EOF_ && wlink._rxbuf.length() != 0)
            err = io::ErrUnexpectedEOF;
        return make_tuple("", err);
    }
}

// from_string parses string into rxPkt.
error rxPkt::from_string(const string &rx) {
    rxPkt& pkt = *this;

    // <stream> ... \n
    auto sp = rx.find(' ');
    if (sp == string::npos)
        return fmt::errorf("invalid pkt: no SP");
    string sid  = rx.substr(0, sp);
    string smsg = rx.substr(sp+1);

    sscanf(sid.c_str(), "%" SCNu64, &pkt.stream);
    if (std::to_string(pkt.stream) != sid)
        return fmt::errorf("invalid pkt: invalid stream ID");

    auto msglen = smsg.length();
    if (msglen > ARRAY_SIZE(pkt.data))
        return fmt::errorf("invalid pkt: len(msg) > %zu", ARRAY_SIZE(pkt.data));

    memcpy(pkt.data, smsg.c_str(), msglen);
    pkt.datalen = msglen;
    return nil;
}

// to_string converts rxPkt data into string.
string rxPkt::to_string() const {
    const rxPkt& pkt = *this;
    return string(pkt.data, pkt.datalen);
}

// ---- WCFS raw file access ----

// _path returns path for object on wcfs.
// - str:        wcfs root + obj;
string WCFS::_path(const string &obj) {
    WCFS *wc = this;

    return wc->mountpoint + "/" + obj;
}

tuple<os::File, error> WCFS::_open(const string &path, int flags) {
    WCFS *wc = this;
    string path_ = wc->_path(path);
    return os::open(path_, flags);
}


// ---- misc ----

static string h(uint64_t v) {
    return fmt::sprintf("%016lx", v);
}

// map_zero_ro mmaps read-only zeros into [addr +size) so that region as all zeros.
// created mapping, even after it is accessed, does not consume memory.
static error _mmap_zero_into_ro(void *addr, size_t size);
static error mmap_zero_into_ro(void *addr, size_t size) {
    error err = _mmap_zero_into_ro(addr, size);
    if (err != nil)
        err = fmt::errorf("mmap zero"); // XXX -> xerr::contextf?
    return err;
}
static error _mmap_zero_into_ro(void *addr, size_t size) {
    // mmap /dev/zero with MAP_NORESERVE and MAP_SHARED
    // this way the mapping will be able to be read, but no memory will be allocated to keep it.
    os::File z;
    error err;
    tie(z, err) = os::open("/dev/zero");
    if (err != nil)
        return err;
    defer([&]() {
        z.close();
    });
    err = mm::map_into(addr, size, PROT_READ, MAP_SHARED | MAP_NORESERVE, z, 0);
    if (err != nil)
        return err;
    return nil;
}
